# -*- coding: utf-8 -*-
from base64 import urlsafe_b64encode
import rpc
import conf
import zlib

UNDEFINED_KEY = "?"

class PutExtra(object):
	callback_params = None
	bucket = None
	custom_meta = None
	mime_type = None
	crc32 = ""
	check_crc = 0
	def __init__(self, bucket):
		self.bucket = bucket

def put(uptoken, key, data, extra):
	action = ["/rs-put"]
	action.append(urlsafe_b64encode("%s:%s" % (extra.bucket, key)))
	if extra.mime_type is not None:
		action.append("mimeType/%s" % urlsafe_b64encode(extra.mime_type))

	if extra.custom_meta is not None:
		action.append("meta/%s" % urlsafe_b64encode(extra.custom_meta))

	if extra.check_crc:
		action.append("crc32/%s" % extra.crc32)

	fields = [
		("action", '/'.join(action)),
		("auth", uptoken),
	]
	if extra.callback_params is not None:
		fields.append(("params", extra.callback_params))

	files = [
		("file", key, data)
	]
	return rpc.Client(conf.UP_HOST).call_with_multipart("/upload", fields, files)

def put_file(uptoken, key, localfile, extra):
	f = open(localfile)
	data = f.read()
	f.close()
	if extra.check_crc == 1:
		extra.crc32 = zlib.crc32(data) & 0xFFFFFFFF
	return put(uptoken, key, data, extra)

def get_url(domain, key, dntoken):
	return "%s/%s?token=%s" % (domain, key, dntoken)




# -*- coding: utf-8 -*-
from .. import conf
from .. import rpc


class Client(rpc.Client):
	up_token = None
	
	def __init__(self, up_token, host=None):
		if host is None:
			host = conf.UP_HOST
		self.up_token = up_token
		super(Client, self).__init__(host)

	def round_tripper(self, method, path, body):
		self.set_header("Authorization", "UpToken %s" % self.up_token)
		return super(Client, self).round_tripper(method, path, body)



# -*- coding: utf-8 -*-
import os
import unittest
import string
import random
import zlib
from base64 import urlsafe_b64encode as encode

from qiniu import conf
from qiniu import rs
from qiniu import io

conf.ACCESS_KEY = os.getenv("QINIU_ACCESS_KEY")
conf.SECRET_KEY = os.getenv("QINIU_SECRET_KEY")
bucket_name = os.getenv("QINIU_BUCKET_NAME")

policy = rs.PutPolicy(bucket_name)
extra = io.PutExtra(bucket_name)
extra.mime_type = "text/plain"

def r(length):
	lib = string.ascii_uppercase
	return ''.join([random.choice(lib) for i in range(0, length)])

class TestUp(unittest.TestCase):
	def test(self):
		def test_put():
			key = "test_%s" % r(9)
			params = "op=3"
			data = "hello bubby!"
			extra.check_crc = 2
			extra.crc32 = zlib.crc32(data) & 0xFFFFFFFF
			ret, err = io.put(policy.token(), key, data, extra)
			assert err is None

		def test_put_same_crc():
			key = "test_%s" % r(9)
			params = "op=3"
			data = "hello bubby!"
			extra.check_crc = 2
			ret, err = io.put(policy.token(), key, data, extra)
			assert err is None

		test_put()
		test_put_same_crc()

	def test_put_file(self):
		localfile = "%s" % __file__
		key = "test_%s" % r(9)

		extra.check_crc = 1
		ret, err = io.put_file(policy.token(), key, localfile, extra)
		assert err is None
		assert ret is not None

	def test_put_crc_fail(self):
		key = "test_%s" % r(9)
		params = "op=3"
		data = "hello bubby!"
		extra.check_crc = 2
		extra.crc32 = "wrong crc32"
		ret, err = io.put(policy.token(), key, data, extra)
		assert err is not None


if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
import os
import zlib
from base64 import urlsafe_b64encode

import auth.up
import conf

UNDEFINED_KEY = "?"

_workers = 1
_task_queue_size = _workers * 4
_chunk_size = 256 * 1024
_try_times = 3
_block_size = 4 * 1024 * 1024

class Error(Exception):
	value = None
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return self.value

err_invalid_put_progress = Error("invalid put progress")
err_put_failed = Error("resumable put failed")
err_unmatched_checksum = Error("unmatched checksum")

def setup(chunk_size=0, try_times=0):
	"""
	 * chunk_size      => 默认的Chunk大小，不设定则为256k
	 * try_times       => 默认的尝试次数，不设定则为3
	"""
	global _chunk_size, _try_times

	if chunk_size == 0:
		chunk_size = 1 << 18

	if try_times == 0:
		try_times = 3

	_chunk_size, _try_times = chunk_size, try_times

# ----------------------------------------------------------
def gen_crc32(data):
	return zlib.crc32(data) & 0xffffffff

class PutExtra(object):
	callback_params = None # 当 uptoken 指定了 CallbackUrl，则 CallbackParams 必须非空
	bucket = None          # 当前是必选项，但未来会去掉
	custom_meta = None     # 可选。用户自定义 Meta，不能超过 256 字节
	mimetype = None        # 可选。在 uptoken 没有指定 DetectMime 时，用户客户端可自己指定 MimeType
	chunk_size = None      # 可选。每次上传的Chunk大小
	try_times = None       # 可选。尝试次数
	progresses = None      # 可选。上传进度
	notify = lambda self, idx, size, ret: None # 可选。进度提示
	notify_err = lambda self, idx, size, err: None

	def __init__(self, bucket):
		self.bucket = bucket

def put_file(uptoken, key, localfile, extra):
	""" 上传文件 """
	f = open(localfile, "rb")
	statinfo = os.stat(localfile)
	ret = put(uptoken, key, f, statinfo.st_size, extra)
	f.close()
	return ret

def put(uptoken, key, f, fsize, extra):
	""" 上传二进制流, 通过将data "切片" 分段上传 """
	if not isinstance(extra, PutExtra):
		print("extra must the instance of PutExtra")
		return

	block_cnt = block_count(fsize)
	if extra.progresses is None:
		extra.progresses = [None for i in xrange(0, block_cnt)]
	else:
		if not len(extra.progresses) == block_cnt:
			return None, err_invalid_put_progress

	if extra.try_times is None:
		extra.try_times = _try_times

	if extra.chunk_size is None:
		extra.chunk_size = _chunk_size

	client = auth.up.Client(uptoken)
	for i in xrange(0, block_cnt):
		try_time = extra.try_times
		read_length = _block_size
		if (i+1)*_block_size > fsize:
			read_length = fsize - i*_block_size
		data_slice = f.read(read_length)
		while True:
			err = resumable_block_put(client, data_slice, i, extra)
			if err is None:
				break

			try_time -= 1
			if try_time <= 0:
				return None, err_put_failed
			print err, ".. retry"

	return mkfile(client, key, fsize, extra)

# ----------------------------------------------------------

def resumable_block_put(client, block, index, extra):
	block_size = len(block)

	if extra.progresses[index] is None or "ctx" not in extra.progresses[index]:
		end_pos = extra.chunk_size-1
		if block_size < extra.chunk_size:
			end_pos = block_size-1
		chunk = block[: end_pos]
		crc32 = gen_crc32(chunk)
		chunk = bytearray(chunk)
		extra.progresses[index], err = mkblock(client, block_size, chunk)
		if not extra.progresses[index]["crc32"] == crc32:
			return err_unmatched_checksum
		if err is not None:
			extra.notify_err(index, end_pos + 1, err)
			return err
		extra.notify(index, end_pos + 1, extra.progresses[index])

	while extra.progresses[index]["offset"] < block_size:
		offset = extra.progresses[index]["offset"]
		chunk = block[offset: offset+extra.chunk_size-1]
		crc32 = gen_crc32(chunk)
		chunk = bytearray(chunk)
		extra.progresses[index], err = putblock(client, extra.progresses[index], chunk)
		if not extra.progresses[index]["crc32"] == crc32:
			return err_unmatched_checksum
		if err is not None:
			extra.notify_err(index, len(chunk), err)
			return err
		extra.notify(index, len(chunk), extra.progresses[index])

def block_count(size):
	global _block_size
	return size / _block_size + 1

def mkblock(client, block_size, first_chunk):
	url = "http://%s/mkblk/%s" % (conf.UP_HOST, block_size)
	content_type = "application/octet-stream"
	return client.call_with(url, first_chunk, content_type, len(first_chunk))

def putblock(client, block_ret, chunk):
	url = "%s/bput/%s/%s" % (block_ret["host"], block_ret["ctx"], block_ret["offset"])
	content_type = "application/octet-stream"
	return client.call_with(url, chunk, content_type, len(chunk))

def mkfile(client, key, fsize, extra):
	encoded_entry = urlsafe_b64encode("%s:%s" % (extra.bucket, key))
	url = ["http://%s/rs-mkfile/%s/fsize/%s" % (conf.UP_HOST, encoded_entry, fsize)]

	if extra.mimetype:
		url.append("mimeType/%s" % urlsafe_b64encode(extra.mimetype))

	if extra.custom_meta:
		url.append("meta/%s" % urlsafe_b64encode(extra.custom_meta))

	if extra.callback_params:
		url.append("params/%s" % urlsafe_b64encode(extra.callback_params))

	url = "/".join(url)
	body = ",".join([i["ctx"] for i in extra.progresses])
	return client.call_with(url, body, "text/plain", len(body))



# -*- coding: utf-8 -*-
import unittest
import os
import json
from base64 import urlsafe_b64decode as decode
from base64 import urlsafe_b64encode as encode
from hashlib import sha1
import hmac
import urllib

from qiniu import conf
from qiniu import rpc
from qiniu import rs

conf.ACCESS_KEY = os.getenv("QINIU_ACCESS_KEY")
conf.SECRET_KEY = os.getenv("QINIU_SECRET_KEY")
bucket_name = os.getenv("QINIU_BUCKET_NAME")
domain = os.getenv("QINIU_DOMAIN")
key = os.getenv("QINIU_PIC_KEY")

class TestToken(unittest.TestCase):
	def test_put_policy(self):
		policy = rs.PutPolicy(bucket_name)
		policy.endUser = "hello!"
		tokens = policy.token().split(':')
		self.assertEqual(conf.ACCESS_KEY, tokens[0])
		data = json.loads(decode(tokens[2]))
		self.assertEqual(data["scope"], bucket_name)
		self.assertEqual(data["endUser"], policy.endUser)

		new_hmac = encode(hmac.new(conf.SECRET_KEY, tokens[2], sha1).digest())
		self.assertEqual(new_hmac, tokens[1])

	def test_get_policy(self):
		base_url = rs.make_base_url(domain, key)
		policy = rs.GetPolicy()
		private_url = policy.make_request(base_url)

		f = urllib.urlopen(private_url)
		body = f.read()
		f.close()
		self.assertEqual(len(body)>100, True)


if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
import os
import unittest
import zlib
import urllib
import tempfile
import shutil

from qiniu import conf
from qiniu.auth import up
from qiniu import resumable_io
from qiniu import rs

bucket = os.getenv("QINIU_BUCKET_NAME")
conf.ACCESS_KEY = os.getenv("QINIU_ACCESS_KEY")
conf.SECRET_KEY = os.getenv("QINIU_SECRET_KEY")

class TestBlock(unittest.TestCase):
	def test_block(self):
		policy = rs.PutPolicy(bucket)
		uptoken = policy.token()
		client = up.Client(uptoken)

		rets = [0, 0]
		data_slice_2 = "\nbye!"
		ret, err = resumable_io.mkblock(client, len(data_slice_2), data_slice_2)
		assert err is None, err 
		self.assertEqual(ret["crc32"], zlib.crc32(data_slice_2))

		extra = resumable_io.PutExtra(bucket)
		extra.mimetype = "text/plain"
		extra.progresses = [ret]
		lens = 0
		for i in xrange(0, len(extra.progresses)):
			lens += extra.progresses[i]["offset"]

		key = "sdk_py_resumable_block_4"
		ret, err = resumable_io.mkfile(client, key, lens, extra)
		assert err is None, err
		self.assertEqual(ret["hash"], "FtCFo0mQugW98uaPYgr54Vb1QsO0", "hash not match")
		rs.Client().delete(bucket, key)
	
	def test_put(self):
		src = urllib.urlopen("http://cheneya.qiniudn.com/hello_jpg")
		dst = tempfile.NamedTemporaryFile()
		shutil.copyfileobj(src, dst)
		src.close()

		policy = rs.PutPolicy(bucket)
		extra = resumable_io.PutExtra(bucket)
		extra.bucket = bucket
		key = "sdk_py_resumable_block_5"
		localfile = dst.name
		ret, err = resumable_io.put_file(policy.token(), key, localfile, extra)
		dst.close()

		assert err is None, err
		self.assertEqual(ret["hash"], "FnyTMUqPNRTdk1Wou7oLqDHkBm_p", "hash not match")
		rs.Client().delete(bucket, key)
			

if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
import httplib
import json
import conf

class Client(object):
	_conn = None
	_header = None
	def __init__(self, host):
		self._conn = httplib.HTTPConnection(host)
		self._header = {}

	def round_tripper(self, method, path, body):
		self._conn.request(method, path, body, self._header)
		resp = self._conn.getresponse()
		return resp

	def call(self, path):
		return self.call_with(path, None)

	def call_with(self, path, body, content_type=None, content_length=None):
		ret = None

		self.set_header("User-Agent", conf.USER_AGENT)
		if content_type is not None:
			self.set_header("Content-Type", content_type)
		
		if content_length is not None:
			self.set_header("Content-Length", content_length)

		resp = self.round_tripper("POST", path, body)
		try:
			ret = resp.read()
			ret = json.loads(ret)
		except IOError, e:
			return None, e
		except ValueError:
			pass
			
		if resp.status / 100 != 2:
			err_msg = ret if "error" not in ret else ret["error"]
			detail = resp.getheader("x-log", None)
			if detail is not None:
				err_msg += ", detail:%s" % detail
				
			return None, err_msg
		
		return ret, None

	def call_with_multipart(self, path, fields=None, files=None):
		"""
		 *  fields => [(key, value)]
		 *  files => [(key, filename, value)]
		"""
		content_type, body = self.encode_multipart_formdata(fields, files)
		return self.call_with(path, body, content_type, len(body))

	def call_with_form(self, path, ops):
		"""
		 * ops => {"key": value/list()}
		"""
		
		body = []
		for i in ops:
			if isinstance(ops[i], (list, tuple)):
				data = ('&%s=' % i).join(ops[i])
			else:
				data = ops[i]
			
			body.append('%s=%s' % (i, data))
		body = '&'.join(body)
		
		content_type = "application/x-www-form-urlencoded"
		return self.call_with(path, body, content_type, len(body))

	def set_header(self, field, value):
		self._header[field] = value

	def set_headers(self, headers):
		self._header.update(headers)

	def encode_multipart_formdata(self, fields, files):
		"""
		 *  fields => [(key, value)]
		 *  files => [(key, filename, value)]
		 *  return content_type, body
		"""
		if files is None:
			files = []
		if fields is None:
			fields = []

		BOUNDARY = '----------ThIs_Is_tHe_bouNdaRY_$'
		CRLF = '\r\n'
		L = []
		for (key, value) in fields:
			L.append('--' + BOUNDARY)
			L.append('Content-Disposition: form-data; name="%s"' % key)
			L.append('')
			L.append(value)
		for (key, filename, value) in files:
			L.append('--' + BOUNDARY)
			disposition = "Content-Disposition: form-data;"
			L.append('%s name="%s"; filename="%s"' % (disposition, key, filename))
			L.append('Content-Type: application/octet-stream')
			L.append('')
			L.append(value)
		L.append('--' + BOUNDARY + '--')
		L.append('')
		body = CRLF.join(L)
		content_type = 'multipart/form-data; boundary=%s' % BOUNDARY
		return content_type, body



# -*- coding:utf-8 -*-
import json

class Exif(object):
	def make_request(self, url):
		return '%s?exif' % url


class ImageView(object):
	mode = 1 # 1或2
	width = None # width 默认为0，表示不限定宽度
	height = None
	quality = None # 图片质量, 1-100
	format = None # 输出格式, jpg, gif, png, tif 等图片格式

	def make_request(self, url):
		target = []
		target.append('%s' % self.mode)
		
		if self.width is not None:
			target.append("w/%s" % self.width)

		if self.height is not None:
			target.append("h/%s" % self.height)

		if self.quality is not None:
			target.append("q/%s" % self.quality)

		if self.format is not None:
			target.append("format/%s" % self.format)

		return "%s?imageView/%s" % (url, '/'.join(target))


class ImageInfo(object):
	def make_request(self, url):
		return '%s?imageInfo' % url



# -*- coding: utf-8 -*-
import auth.digest
import conf
import urllib

EOF = 'EOF'


class Client(object):
	conn = None
	def __init__(self, mac=None):
		if mac is None:
			mac = auth.digest.Mac()
		self.conn = auth.digest.Client(host=conf.RSF_HOST, mac=mac)
		
	def list_prefix(self, bucket, prefix=None, marker=None, limit=None):
		'''前缀查询:
		 * bucket => str
		 * prefix => str
		 * marker => str
		 * limit => int
		 * return ret => {'items': items, 'marker': markerOut}, err => str

		1. 首次请求 marker = None
		2. 无论 err 值如何，均应该先看 ret.get('items') 是否有内容
		3. 如果后续没有更多数据，err 返回 EOF，markerOut 返回 None（但不通过该特征来判断是否结束） 
		'''
		ops = {
			'bucket': bucket,
		}
		if marker is not None:
			ops['marker'] = marker
		if limit is not None:
			ops['limit'] = limit
		if prefix is not None:
			ops['prefix'] = prefix
		url = '%s?%s' % ('/list', urllib.urlencode(ops))
		ret, err = self.conn.call_with(url, body=None, content_type='application/x-www-form-urlencoded')
		if not ret.get('marker'):
			err = EOF
		return ret, err



# -*- coding: utf-8 -*-
import json
import time
import urllib

from ..auth import digest

class PutPolicy(object):
	scope = None             # 可以是 bucketName 或者 bucketName:key
	expires = 3600           # 默认是 3600 秒
	callbackUrl = None
	callbackBody = None
	returnUrl = None
	returnBody = None
	endUser = None
	asyncOps = None         

	def __init__(self, scope):
		self.scope = scope

	def token(self, mac=None):
		if mac is None:
			mac = digest.Mac()
		token = dict(
			scope = self.scope,
			deadline = int(time.time()) + self.expires,
		)

		if self.callbackUrl is not None:
			token["callbackUrl"] = self.callbackUrl

		if self.callbackBody is not None:
			token["callbackBody"] = self.callbackBody

		if self.returnUrl is not None:
			token["returnUrl"] = self.returnUrl

		if self.returnBody is not None:
			token["returnBody"] = self.returnBody

		if self.endUser is not None:
			token["endUser"] = self.endUser

		if self.asyncOps is not None:
			token["asyncOps"] = self.asyncOps
		
		b = json.dumps(token, separators=(',',':'))
		return mac.sign_with_data(b)

class GetPolicy(object):
	expires = 3600
	def __init__(self):
		pass
	
	def make_request(self, base_url, mac=None):
		'''
		 *  return private_url
		'''
		if mac is None:
			mac = digest.Mac()

		deadline = int(time.time()) + self.expires
		if '?' in base_url:
			base_url += '&'
		else:
			base_url += '?'
		base_url = '%se=%s' % (base_url, str(deadline))

		token = mac.sign(base_url)
		return '%s&token=%s' % (base_url, token)


def make_base_url(domain, key):
	'''
	 * domain => str
	 * key => str
	 * return base_url
	'''
	return 'http://%s/%s' % (domain, urllib.quote(key))



# -*- coding: utf-8 -*-
import unittest

from qiniu import rpc
from qiniu import conf

def round_tripper(client, method, path, body):
	pass

class ClsTestClient(rpc.Client):
	def round_tripper(self, method, path, body):
		round_tripper(self, method, path, body)
		return super(ClsTestClient, self).round_tripper(method, path, body)

client = ClsTestClient(conf.RS_HOST)

class TestClient(unittest.TestCase):
	def test_call(self):
		global round_tripper
		
		def tripper(client, method, path, body):
			self.assertEqual(path, "/hello")
			assert body is None
		
		round_tripper = tripper
		client.call("/hello")

	def test_call_with(self):
		global round_tripper
		def tripper(client, method, path, body):
			self.assertEqual(body, "body")
		
		round_tripper = tripper
		client.call_with("/hello", "body")

	def test_call_with_multipart(self):
		global round_tripper
		def tripper(client, method, path, body):
			self.assertEqual(len(body), client._header["Content-Length"])
			target_type = "multipart/form-data"
			self.assertTrue(client._header["Content-Type"].startswith(target_type))
			start_index = client._header["Content-Type"].find("boundary")
			boundary = client._header["Content-Type"][start_index + 9: ]
			dispostion = 'Content-Disposition: form-data; name="auth"'
			tpl = "--%s\r\n%s\r\n\r\n%s\r\n--%s--\r\n" % (boundary, dispostion,
					"auth_string", boundary)
			self.assertEqual(tpl, body)
		
		round_tripper = tripper
		client.call_with_multipart("/hello", fields=[("auth", "auth_string")])

	def test_call_with_form(self):
		global round_tripper
		def tripper(client, method, path, body):
			self.assertEqual(body, "action=a&op=a&op=b")
			target_type = "application/x-www-form-urlencoded"
			self.assertEqual(client._header["Content-Type"], target_type)
			self.assertEqual(client._header["Content-Length"], len(body))
		
		round_tripper = tripper
		client.call_with_form("/hello", dict(op=["a", "b"], action="a"))

if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
import unittest
import os
import random
import string

from qiniu import rs
from qiniu import conf

def r(length):
	lib = string.ascii_uppercase
	return ''.join([random.choice(lib) for i in range(0, length)])

conf.ACCESS_KEY = os.getenv("QINIU_ACCESS_KEY")
conf.SECRET_KEY = os.getenv("QINIU_SECRET_KEY")
key = os.getenv("QINIU_PIC_KEY")
bucket_name = os.getenv("QINIU_BUCKET_NAME")
noexist_key = os.getenv("QINIU_NOEXIST_PIC_KEY")
key2 = "rs_demo_test_key_1_" + r(5)
key3 = "rs_demo_test_key_2_" + r(5)
key4 = "rs_demo_test_key_3_" + r(5)

class TestRs(unittest.TestCase):
	def test_stat(self):
		r = rs.Client()
		ret, err = r.stat(bucket_name, key)
		assert err is None
		assert ret is not None
		
		# error
		_, err = r.stat(bucket_name, noexist_key)
		assert err is not None
	
	def test_delete_move_copy(self):
		r = rs.Client()
		r.delete(bucket_name, key2)
		r.delete(bucket_name, key3)
		
		ret, err = r.copy(bucket_name, key, bucket_name, key2)
		assert err is None, err
		
		ret, err = r.move(bucket_name, key2, bucket_name, key3)
		assert err is None, err
		
		ret, err = r.delete(bucket_name, key3)
		assert err is None, err
		
		# error
		_, err = r.delete(bucket_name, key2)
		assert err is not None
		
		_, err = r.delete(bucket_name, key3)
		assert err is not None

	def test_batch_stat(self):
		r = rs.Client()
		entries = [
			rs.EntryPath(bucket_name, key),
			rs.EntryPath(bucket_name, key2),
		]
		ret, err = r.batch_stat(entries)
		assert err is None
		self.assertEqual(ret[0]["code"], 200)
		self.assertEqual(ret[1]["code"], 612)

	def test_batch_delete_move_copy(self):
		r = rs.Client()
		e1 = rs.EntryPath(bucket_name, key)
		e2 = rs.EntryPath(bucket_name, key2)
		e3 = rs.EntryPath(bucket_name, key3)
		e4 = rs.EntryPath(bucket_name, key4)
		r.batch_delete([e2, e3, e4])
		
		# copy
		entries = [
			rs.EntryPathPair(e1, e2),
			rs.EntryPathPair(e1, e3),
		]
		ret, err = r.batch_copy(entries)
		assert err is None
		self.assertEqual(ret[0]["code"], 200)
		self.assertEqual(ret[1]["code"], 200)
		
		ret, err = r.batch_move([rs.EntryPathPair(e2, e4)])
		assert err is None
		self.assertEqual(ret[0]["code"], 200)
		
		ret, err = r.batch_delete([e3, e4])
		assert err is None
		self.assertEqual(ret[0]["code"], 200)
		
		r.batch_delete([e2, e3, e4])

if __name__ == "__main__":
	unittest.main()
